/*
 * $Id: gmx_rdf.c,v 1.12.2.3 2006/03/23 07:43:05 spoel Exp $
 * 
 *                This source code is part of
 * 
 *                 G   R   O   M   A   C   S
 * 
 *          GROningen MAchine for Chemical Simulations
 * 
 *                        VERSION 3.2.0
 * Written by David van der Spoel, Erik Lindahl, Berk Hess, and others.
 * Copyright (c) 1991-2000, University of Groningen, The Netherlands.
 * Copyright (c) 2001-2004, The GROMACS development team,
 * check out http://www.gromacs.org for more information.

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * If you want to redistribute modifications, please consider that
 * scientific software is very special. Version control is crucial -
 * bugs must be traceable. We will be happy to consider code for
 * inclusion in the official distribution, but derived work must not
 * be called official GROMACS. Details are found in the README & COPYING
 * files - if they are missing, get the official version at www.gromacs.org.
 * 
 * To help us fund GROMACS development, we humbly ask that you cite
 * the papers on the package - you can find them in the top README file.
 * 
 * For more info, check our website at http://www.gromacs.org
 * 
 * And Hey:
 * Green Red Orange Magenta Azure Cyan Skyblue
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <math.h>
#include <ctype.h>
#include "string2.h"
#include "sysstuff.h"
#include "typedefs.h"
#include "macros.h"
#include "vec.h"
#include "pbc.h"
#include "rmpbc.h"
#include "xvgr.h"
#include "copyrite.h"
#include "futil.h"
#include "statutil.h"
#include "tpxio.h"
#include "index.h"
#include "smalloc.h"
#include "nrnb.h"
#include "shift_util.h"

typedef struct
{
  const char   *Label;
  real        a[4], b[4], c;
} t_CM_table;

/*
 * 
 * f0[k] = c + [SUM a_i*EXP(-b_i*(k^2)) ]
 *             i=1,4
 */

const t_CM_table CM_t[] =
{
  { "H",    { 0.489918, 0.262003, 0.196767, 0.049879 },
    { 20.6593, 7.74039, 49.5519, 2.20159 },
    0.001305 },
 { "HO",    { 0.489918, 0.262003, 0.196767, 0.049879 },
   { 20.6593, 7.74039, 49.5519, 2.20159 },
   0.001305 },
  { "HW",    { 0.489918, 0.262003, 0.196767, 0.049879 },
    { 20.6593, 7.74039, 49.5519, 2.20159 },
    0.001305 },
  { "C",  { 2.26069, 1.56165, 1.05075, 0.839259 },
    { 22.6907, 0.656665, 9.75618, 55.5949 },
    0.286977 },
  { "CB",  { 2.26069, 1.56165, 1.05075, 0.839259 },
    { 22.6907, 0.656665, 9.75618, 55.5949 },
    0.286977 },
  { "CS1", { 0., 0., 0., 0.},{ 0., 0., 0., 0.}, 0.}, 
  { "CS2", { 0., 0., 0., 0.},{ 0., 0., 0., 0.}, 0.},
  { "N",     { 12.2126, 3.13220, 2.01250, 1.16630 },     
    { 0.005700, 9.89330, 28.9975, 0.582600 },
    -11.529 },
  { "O",      { 3.04850, 2.28680, 1.54630, 0.867000 },  
    { 13.2771, 5.70110, 0.323900, 32.9089 },
    0.250800 },
  { "OW",     { 3.04850, 2.28680, 1.54630, 0.867000 }, 
    { 13.2771, 5.70110, 0.323900, 32.9089 },
    0.250800 },
  { "OWT3",   { 3.04850, 2.28680, 1.54630, 0.867000 },   
    { 13.2771, 5.70110, 0.323900, 32.9089 },
    0.250800 },
  { "OA",     { 3.04850, 2.28680, 1.54630, 0.867000 },  
    { 13.2771, 5.70110, 0.323900, 32.9089 },
    0.250800 },
  { "OS",     { 3.04850, 2.28680, 1.54630, 0.867000 }, 
    { 13.2771, 5.70110, 0.323900, 32.9089 },
    0.250800 },
  { "OSE",    { 3.04850, 2.28680, 1.54630, 0.867000 },
    { 13.2771, 5.70110, 0.323900, 32.9089 },
    0.250800 },
  { "Na",  { 3.25650, 3.93620, 1.39980, 1.00320 },       /*  Na 1+ */
    { 2.66710, 6.11530, 0.200100, 14.0390 }, 
    0.404000 },
  { "CH3", { 0., 0., 0., 0.},{ 0., 0., 0., 0.}, 0.},    
  { "CH2", { 0., 0., 0., 0.},{ 0., 0., 0., 0.}, 0.},    
  { "CH1", { 0., 0., 0., 0.},{ 0., 0., 0., 0.}, 0.},   
};

typedef struct
{
  const char   *Label;
  real         m;
} t_CM_list;

const t_CM_list CM_l[] =
{
  { "LO",     15.9994 },
  { "LOM",    15.9994 },
  { "LNL",    14.0067 },
  { "LC",     12.0110 },
  { "LH1",    13.0190 },
  { "LH2",    14.0270 },
  { "LP",     30.9738 },
  { "LOS",    15.9994 },
  { "LP2",    14.0270 },
  { "LP3",    15.0350 },
  { "LC3",    15.0350 },
  { "LC2",    14.0270 }
};

static void check_box_c(matrix box)
{
  if (fabs(box[ZZ][XX]) > GMX_REAL_EPS*box[ZZ][ZZ] ||
      fabs(box[ZZ][YY]) > GMX_REAL_EPS*box[ZZ][ZZ])
    gmx_fatal(FARGS,
	      "The last box vector is not parallel to the z-axis: %f %f %f",
	      box[ZZ][XX],box[ZZ][YY],box[ZZ][ZZ]);
}

static void do_rdf(char *fnNDX,char *fnTPS,char *fnTRX,char *fnRDF,
                   bool bPBC,bool bCentroid,bool bTrans,
		   real cutoff,real binwidth,int ng)
{
  FILE       *fp;
  int        status;
  char       outf1[STRLEN],outf2[STRLEN];
  char       title[STRLEN];
  int        g,natoms,i,j,k,nbin,n,nframes;
  int        *count_i;
  char       **grpname;
  int        *isize,*nrdf,*isize2;
  atom_id    **index,**index2;
  real       t,rmax2,cut2,r,r2,invbinw,normfac;
  real       segvol,spherevol,prev_spherevol,**rdf;
  rvec       *x,xcom,dx,xi,xj,**xcm;
  real       *inv_segvol,invvol;
  matrix     box,box_pbc;
  atom_id    ix,jx;
  t_topology top;
  bool       bTop;
  t_pbc      pbc;
  real       zmid,rho_i,mass,tmass;
  int        ni,resnr,last_resnr,mol,nUpper;
  bool       bGhost,**bUpper,bLeaflet;
  int        nclose=0;

  
  if (fnTPS) {
    bTop=read_tps_conf(fnTPS,title,&top,&x,NULL,box,TRUE);
    mk_single_top(&top);
  }
  snew(grpname,ng+1);
  snew(isize,ng+1);
  snew(isize2,ng+1);
  snew(index,ng+1);
  snew(index2,ng+1);
  snew(nrdf,ng);
  fprintf(stderr,"\nSelect a reference group and %d group%s\n",
	  ng,ng==1?"":"s");
  if (fnTPS)
    get_index(&top.atoms,fnNDX,ng+1,isize,index,grpname);
  else
    rd_index(fnNDX,ng+1,isize,index,grpname);
  for (g=0; g<ng+1; g++)
    snew(index2[g],isize[g]);
  
  natoms=read_first_x(&status,fnTRX,&t,&x,box);
  if ( !natoms )
    gmx_fatal(FARGS,"Could not read coordinates from statusfile\n");
  if (fnTPS)
    /* check with topology */
    if ( natoms > top.atoms.nr ) 
      gmx_fatal(FARGS,"Trajectory (%d atoms) does not match topology (%d atoms)",
		  natoms,top.atoms.nr);
  /* check with index groups */
  for (i=0; i<=ng; i++)
    for (j=0; j<isize[i]; j++)
      if ( index[i][j] >= natoms )
	gmx_fatal(FARGS,"Atom index (%d) in index group %s (%d atoms) larger "
		    "than number of atoms in trajectory (%d atoms)",
		    index[i][j],grpname[i],isize[i],natoms);
  
  /* initialize some handy things */
  copy_mat(box,box_pbc);
  check_box_c(box);
  /* Make sure the z-height does not influence the cut-off */
  box_pbc[ZZ][ZZ] = 2*max(box[XX][XX],box[YY][YY]);
  rmax2   = 0.99*0.99*max_cutoff2(box_pbc);
  nbin    = (int)(sqrt(rmax2) / binwidth);
  invbinw = 1.0 / binwidth;
  cut2    = sqr(cutoff);
#define DEBUG 100

  snew(rdf,ng);
  for(g=0; g<ng; g++) {
    snew(rdf[g],nbin+1);
  }
  
  /* Calculate volume (actually area) of bin segments */
  snew(inv_segvol,nbin);
  snew(count_i,nbin);
  prev_spherevol=0;
  for(i=0; (i<nbin); i++) {
    r = (i+1)*binwidth;
    spherevol=M_PI*r*r;
    segvol=spherevol-prev_spherevol;
    inv_segvol[i]=1.0/segvol;
    prev_spherevol=spherevol;
  }
  
  snew(xcm,ng+1);
  snew(bUpper,ng+1);
  nframes = 0;
  do {

    ni = 0; zmid = 0.0; /* estimate bilayer midpoint from ref grp */
    for (i=0; i<isize[0]; i++) {
      ix = index[0][i];
      if (x[ix][XX] < 50.0) {
        zmid += x[ix][ZZ];
        ni++;
      }
    }
    zmid /= (real)ni;
    if (nframes%DEBUG == 0)
      printf("\nFRAME %d bilayer midpoint %f\n",nframes,zmid);

    /* determine number of molecules in each grp */
    for (g=0; g<ng+1; g++) {
      mol = 0; last_resnr = -1;
      for (i=0; i<isize[g]; i++) {
        ix = index[g][i];
        resnr = top.atoms.atom[ix].resnr;
        if (resnr != last_resnr && x[ix][XX] < 50.0) {
          mol++;
        }
        last_resnr = resnr;
      }
      isize2[g] = mol;
      if (nframes%DEBUG == 0)
        printf("group %d n %d\n",g,isize2[g]);
    }

    /* Make xcm coords and index2 */
    for (g=0; g<ng+1; g++) {
      srenew(xcm[g],isize2[g]);
      srenew(index2[g],isize2[g]);
      mol = 0; last_resnr = -1;
      bGhost = TRUE;

      for (i=0; i<isize[g]; i++) {
        ix = index[g][i];
        resnr = top.atoms.atom[ix].resnr;
        if(bCentroid)
          mass = 1.0;
        else
          mass  = top.atoms.atom[ix].m;
        if(resnr != last_resnr) {
          if (!bGhost) {
            for(j=0; j<DIM; j++)
              xcm[g][mol][j] = xcom[j]/tmass;
            index2[g][mol] = resnr;
//            printf("g %d mol %d tmass %f x %f\n",g,mol,tmass,xcm[g][mol][XX]);
            mol++;
          }
          clear_rvec(xcom);
          tmass = 0;
          bGhost = (x[ix][XX]>50.0);
        }
        for(j=0; j<DIM; j++)
          xcom[j] += mass*x[ix][j];
        tmass+=mass;
        last_resnr = resnr;
      }

      if (!bGhost) {
        for(j=0; j<DIM; j++)
          xcm[g][mol][j] = xcom[j]/tmass;
        index2[g][mol] = resnr;
//        printf("g %d mol %d tmass %f x %f\n",g,mol,tmass,xcm[g][mol][XX]);
      }
    }

    for (g=0; g<ng+1; g++) {  /* leaflet classification */
      srenew(bUpper[g],isize2[g]);
      for (i=0; i<isize2[g]; i++) {
        bUpper[g][i] = (xcm[g][i][ZZ] > zmid);
      }
    }

    for (g=0; g<ng+1; g++) {
      nUpper = 0;
      for (i=0; i<isize2[g]; i++) {
        ix = index2[g][i];
        nUpper += bUpper[g][i];
//        if (nframes%(10*DEBUG) == 0)
//          printf("g %d mol %d resnr %d x %.2f %.2f %.2f bUpper %d\n",
//                g,i,ix,xcm[g][i][XX],xcm[g][i][YY],xcm[g][i][ZZ],bUpper[g][i]);
      }
      if (nframes%DEBUG == 0)
        printf("g %d nUpper %d nLower %d\n",g,nUpper,isize2[g]-nUpper);
    }

    /* Must init pbc every step because of pressure coupling */
    copy_mat(box,box_pbc);
    if (bPBC) {
      rm_pbc(&top.idef,natoms,box,x,x);
      check_box_c(box);
      box_pbc[ZZ][ZZ] = 2*max(box[XX][XX],box[YY][YY]);
      set_pbc(&pbc,box_pbc);

      /* Set z-size to 1 so we get the surface iso the volume */
      box_pbc[ZZ][ZZ] = 1;
    }
    invvol = 1/det(box_pbc);
    if (nframes%DEBUG == 0)
      printf("vol %f invvol %f\n",1.0/invvol,invvol);

    for(g=0; g<ng; g++) {
      for(i=0; i<isize2[0]; i++) {
        ix = index2[0][i];
	copy_rvec(xcm[0][i],xi);
        ni=0;
        for(j=0; j<nbin; j++)
          count_i[j]=0;
        for(j=0; j<isize2[g+1]; j++) {
          jx = index2[g+1][j];
	  copy_rvec(xcm[g+1][j],xj);
          bLeaflet = (!bTrans) == (bUpper[0][i] == bUpper[g+1][j]);
          if(ix != jx && bLeaflet) {
            if (bPBC)
              pbc_dx(&pbc,xi,xj,dx);
            else
              rvec_sub(xi,xj,dx);
            r2 = dx[XX]*dx[XX] + dx[YY]*dx[YY];
            if (r2>cut2 && r2<=rmax2)
              count_i[(int)(sqrt(r2)*invbinw)]++;  /* IT HAPPENS HERE */
            if (DEBUG == -1 && r2<0.01) {
              nclose++;
              printf("Ooh thats close frame %d nclose %d: %d (%.2f %.2f %.2f) %d (%.2f %.2f %.2f) "
                "r= %f\n",nframes,nclose,ix,xi[XX],xi[YY],xi[ZZ],jx,xj[XX],xj[YY],xj[ZZ],sqrt(r2));
            }
            ni++;
          }
        }
        rho_i = invvol*(real)ni;
        nrdf[g]++;
        for(j=0; j<nbin; j++) {
          rdf[g][j] += count_i[j]*inv_segvol[j]/rho_i;
//          if (nframes%(10*DEBUG) == 0)
//            printf("g %d nrdf %d i %d resnr %d bin %d count %d isegvol %f ni %d rho %f rdf %f\n",
//              g,nrdf[g],i,ix,j,count_i[j],inv_segvol[j],ni,rho_i,rdf[g][j]/(real)nrdf[g]);
        }
      }
    }
    fflush(NULL);
    nframes++;
  } while (read_next_x(status,&t,natoms,x,box));
  fprintf(stderr,"\n");
  
  close_trj(status);
  
  sfree(x);
  
  for(g=0; g<ng; g++) {
    /* We have to normalize by dividing by the number of at-bats */
    normfac = 1.0/(real)nrdf[g];
      
    /* Do the normalization */
    for(i=0; (i<nbin); i++) {
      rdf[g][i] = rdf[g][i]*normfac;
    }
  }
    
  fp=xvgropen(fnRDF,"Radial Distribution","r","");
  if (ng==1)
    fprintf(fp,"@ subtitle \"%s-%s\"\n",grpname[0],grpname[1]);
  else {
    fprintf(fp,"@ subtitle \"reference %s\"\n",grpname[0]);
    xvgr_legend(fp,ng,grpname+1);
  }
  for(i=0; (i<nbin); i++) {
    r = (i+0.5)*binwidth;
    fprintf(fp,"%10g", r);
    for(g=0; g<ng; g++)
      fprintf(fp," %10g",rdf[g][i]);
    fprintf(fp,"\n");
  }
  ffclose(fp);
  
  do_view(fnRDF,"-nxy");

  for(g=0; g<ng; g++)
    sfree(rdf[g]);
  sfree(rdf);
}


int main(int argc,char *argv[])
{
  static char *desc[] = {
    "This computes the lateral center of mass radial distribution function for",
    "lipids embedded within a bilayer. By default, only intra-leaflet correlation",
    "is considered. All ghost atoms (x>50) are excluded automatically.[PAR]",
    "The option [TT]-cut[tt] is meant as an alternative way to avoid",
    "intramolecular peaks in the rdf plot. The option [TT]-centroid[tt]",
    "turns off the mass weighting. This mode will be set automatically if",
    "there no topology file is supplied. The option [TT]-trans[tt]",
    "will only use molecule pairs from opposite leaflets."
  };
  static bool bPBC=TRUE,bCentroid=FALSE,bTrans=FALSE;
  static real cutoff=0,binwidth=0.002;
  static int  ngroups=1;
  t_pargs pa[] = {
    { "-bin",      FALSE, etREAL, {&binwidth},
      "Binwidth (nm)" },
    { "-pbc",      FALSE, etBOOL, {&bPBC},
      "Use periodic boundary conditions for computing distances" },
    { "-trans",    FALSE, etBOOL, {&bTrans},
      "Trans-leaflet correlation instead of intra-leaflet" },
    { "-centroid", FALSE, etBOOL, {&bCentroid},
      "Use molecule centroid instead of center of mass" },
    { "-cut",      FALSE, etREAL, {&cutoff},
      "Shortest distance (nm) to be considered"},
    { "-ng",       FALSE, etINT, {&ngroups},
      "Number of secondary groups to compute RDFs around a central group" }
  };
#define NPA asize(pa)
  char       *fnTPS,*fnNDX;
  
  t_filenm   fnm[] = {
    { efTRX, "-f",  NULL,     ffREAD },
    { efTPS, NULL,  NULL,     ffOPTRD },
    { efNDX, NULL,  NULL,     ffOPTRD },
    { efXVG, "-o",  "rdf",    ffOPTWR },
  };
#define NFILE asize(fnm)
  
  CopyRight(stderr,argv[0]);
  parse_common_args(&argc,argv,PCA_CAN_VIEW | PCA_CAN_TIME | PCA_BE_NICE,
		    NFILE,fnm,NPA,pa,asize(desc),desc,0,NULL);

  fnTPS = ftp2fn_null(efTPS,NFILE,fnm);
  fnNDX = ftp2fn_null(efNDX,NFILE,fnm);
  
  if (!fnTPS && !fnNDX)
    gmx_fatal(FARGS,"Neither index file nor topology file specified\n"
		  "             Nothing to do!");
 
  if(!fnTPS) 
    bCentroid=TRUE;

  do_rdf(fnNDX,fnTPS,ftp2fn(efTRX,NFILE,fnm),
	   opt2fn("-o",NFILE,fnm),
	   bPBC,bCentroid,bTrans,cutoff,binwidth,ngroups);

  thanx(stderr);
  
  return 0;
}
