/*
 * $Id: gmx_density.c,v 1.5.2.1 2005/10/05 21:33:24 lindahl Exp $
 * 
 *                This source code is part of
 * 
 *                 G   R   O   M   A   C   S
 * 
 *          GROningen MAchine for Chemical Simulations
 * 
 *                        VERSION 3.2.0
 * Written by David van der Spoel, Erik Lindahl, Berk Hess, and others.
 * Copyright (c) 1991-2000, University of Groningen, The Netherlands.
 * Copyright (c) 2001-2004, The GROMACS development team,
 * check out http://www.gromacs.org for more information.

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * If you want to redistribute modifications, please consider that
 * scientific software is very special. Version control is crucial -
 * bugs must be traceable. We will be happy to consider code for
 * inclusion in the official distribution, but derived work must not
 * be called official GROMACS. Details are found in the README & COPYING
 * files - if they are missing, get the official version at www.gromacs.org.
 * 
 * To help us fund GROMACS development, we humbly ask that you cite
 * the papers on the package - you can find them in the top README file.
 * 
 * For more info, check our website at http://www.gromacs.org
 * 
 * And Hey:
 * Green Red Orange Magenta Azure Cyan Skyblue
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <math.h>
#include <ctype.h>

#include "sysstuff.h"
#include "string.h"
#include "string2.h"
#include "typedefs.h"
#include "smalloc.h"
#include "macros.h"
#include "gstat.h"
#include "vec.h"
#include "xvgr.h"
#include "pbc.h"
#include "copyrite.h"
#include "futil.h"
#include "statutil.h"
#include "index.h"
#include "tpxio.h"

/****************************************************************************/
/* This program calculates the partial density across the box.              */
/* Peter Tieleman, Mei 1995                                                 */
/****************************************************************************/

void center_coords(t_atoms *atoms,matrix box,rvec x0[],int axis)
{
  int  i,m;
  real tmass,mm;
  rvec com,shift,box_center;
  
  tmass = 0;
  clear_rvec(com);
  for(i=0; (i<atoms->nr); i++) {
    mm     = atoms->atom[i].m;
    tmass += mm;
    for(m=0; (m<DIM); m++) 
      com[m] += mm*x0[i][m];
  }
  for(m=0; (m<DIM); m++) 
    com[m] /= tmass;
  calc_box_center(ecenterDEF,box,box_center);
  rvec_sub(box_center,com,shift);
  shift[axis] -= box_center[axis];
  
  for(i=0; (i<atoms->nr); i++) 
    rvec_dec(x0[i],shift);
}


void calc_thickness(char *fn, atom_id **index, int gnx[], real ***thickness, 
                      real **time, int *nframe, t_topology *top, int axis, 
                      int nr_grps)
{
  int n, i, ai, natoms, status, nt, nb, nc, nr_frames = 0;
  real t, z, zc, zt, zb, thick_t, thick_avg[nr_grps];
  rvec *x0;              /* coordinates without pbc */
  matrix box;            /* box (3x3) */

  *time = NULL;
  snew(*thickness, nr_grps);
  for (n=0; n < nr_grps; n++) {
    thick_avg[n] = 0.0;
    (*thickness)[n] = NULL;
  }

  if ((natoms = read_first_x(&status,fn,&t,&x0,box)) == 0)
    gmx_fatal(FARGS,"Could not read coordinates from statusfile\n");

  do {

    for (n=0; n < nr_grps; n++) {

      /* find the middle first */
      zc = 0.0; nc = 0;
      for (i=0; i < gnx[n]; i++) {
        ai = index[n][i];
        if (strcmp(*(top->atoms.atomtype[ai]),"LP") == 0 && x0[ai][XX]<50.0) {
          zc += x0[ai][axis];
          nc++;
        }
      }
      if (nc == 0) gmx_fatal(FARGS,"No LP atoms found in grp %d.\n",n);
      zc /= (real)nc;
//      printf("Grp %d Middle %d %f\n",n,nc,zc);

      zt = zb = 0.0; nt = nb = 0;
      for (i=0; i < gnx[n]; i++) {
        ai = index[n][i];
        if (strcmp(*(top->atoms.atomtype[ai]),"LP") == 0 && x0[ai][XX]<50.0) {
          z = x0[ai][axis];
          if (z > zc) { zt += z; nt++; }
          else        { zb += z; nb++; }
        }
      }
      zt /= (real)nt;
      zb /= (real)nb;
//      printf("Grp %d Top %d %f Bottom %d %f\n",n,nt,zt,nb,zb);

      thick_t = (zt - zb);
      thick_avg[n] += thick_t;

      if (nr_frames % 1000 == 0)
        srenew((*thickness)[n], nr_frames + 1000);
      (*thickness)[n][nr_frames] = thick_t;

    } /* grps */

    if (nr_frames % 1000 == 0)
      srenew(*time, nr_frames + 1000);
    (*time)[nr_frames] = t;

    nr_frames++;
  } while (read_next_x(status,&t,natoms,x0,box));

  printf("Frames %d Avg-thickness",nr_frames);
  for (n=0; n < nr_grps; n++) {
    thick_avg[n] /= (real)nr_frames;
    printf(" %f",thick_avg[n]);
  }
  printf("\n");

  (*nframe) = nr_frames;

  close_trj(status);
  sfree(x0);  /* free memory used by coordinate array */
}

void calc_density(char *fn, atom_id **index, int gnx[], 
		  real ***slDensity, int *nslices, t_topology *top, 
		  int axis, int nr_grps, real *slWidth, bool bNumber,
		  bool bCount,bool bCenter)
{
  rvec *x0;              /* coordinates without pbc */
  matrix box;            /* box (3x3) */
  int natoms,            /* nr. atoms in trj */
      status,  
      **slCount,         /* nr. of atoms in one slice for a group */
      i,j,n,               /* loop indices */
      teller = 0,      
      ax1=0, ax2=0,
      nr_frames = 0,     /* number of frames */
      slice;             /* current slice */
  real t, 
        z;
  char *buf;             /* for tmp. keeping atomname */

  switch(axis) {
  case 0:
    ax1 = 1; ax2 = 2;
    break;
  case 1:
    ax1 = 0; ax2 = 2;
    break;
  case 2:
    ax1 = 0; ax2 = 1;
    break;
  default:
    gmx_fatal(FARGS,"Invalid axes. Terminating\n");
  }

  if ((natoms = read_first_x(&status,fn,&t,&x0,box)) == 0)
    gmx_fatal(FARGS,"Could not read coordinates from statusfile\n");
  
  if (! *nslices) {
    *nslices = (int)(box[axis][axis] * 10); /* default value */
    fprintf(stderr,"\nDividing the box in %d slices\n",*nslices);
  }
  
  snew(*slDensity, nr_grps);
  for (i = 0; i < nr_grps; i++)
    snew((*slDensity)[i], *nslices);
  
  /*********** Start processing trajectory ***********/
  do {
    rm_pbc(&(top->idef),top->atoms.nr,box,x0,x0);

    if (bCenter)
      center_coords(&top->atoms,box,x0,axis);
    
    *slWidth = box[axis][axis]/(*nslices);
    teller++;
    
    for (n = 0; n < nr_grps; n++) {      
      for (i = 0; i < gnx[n]; i++) {   /* loop over all atoms in index file */
	z = x0[index[n][i]][axis];
	while (z < 0) 
	  z += box[axis][axis];
	while (z > box[axis][axis])
	  z -= box[axis][axis];
      
	/* determine which slice atom is in */
	slice = (int)(z / (*slWidth)); 
	if (bNumber || bCount) {
	  buf = strdup(*(top->atoms.atomname[index[n][i]]));
	  trim(buf);
	  if (buf[0] != 'H')
	    (*slDensity)[n][slice] += top->atoms.atom[index[n][i]].m;
	  free(buf);
	} else
	  (*slDensity)[n][slice] += top->atoms.atom[index[n][i]].m;
      }
    }

    nr_frames++;
  } while (read_next_x(status,&t,natoms,x0,box));
  
  /*********** done with status file **********/
  close_trj(status);
  
  /* slDensity now contains the total mass per slice, summed over all
     frames. Now divide by nr_frames and volume of slice 
     */
  
  fprintf(stderr,"\nRead %d frames from trajectory. Calculating density\n",
	  nr_frames);

  for (n =0; n < nr_grps; n++) {
    for (i = 0; i < *nslices; i++) {
      if (bCount) 
	(*slDensity)[n][i] = (*slDensity)[n][i]/nr_frames;
      else
	(*slDensity)[n][i] = (*slDensity)[n][i] * (*nslices) * 1.66057 /
	(nr_frames * box[axis][axis] * box[ax1][ax1] * box[ax2][ax2]);
    }
  }

  sfree(x0);  /* free memory used by coordinate array */
}

void plot_thickness(real *thickness[], real time[], char *filenm, 
                      int nr_frames, int nr_grps)
{
  FILE  *out;     /* xvgr file with thickness time series   */
  char  title[256]; /* for xvgr title */
  int i,n;

  sprintf(title,"Bilayer thickness");
  out = xvgropen(filenm, title, "Time (ps)","Thickness (nm)");

  for (i=0; i<nr_frames; i++) {
    fprintf(out,"%f",time[i]);
    for (n=0; n<nr_grps; n++)
      fprintf(out," %f",thickness[n][i]);
    fprintf(out,"\n");
  }

  fclose(out);
}

void plot_density(real *slDensity[], char *afile, int nslices,
		  int nr_grps, char *grpname[], real slWidth, 
		  bool bNumber, bool bCount,
		  bool bSymmetrize)
{
  FILE  *den;     /* xvgr file with density   */
  char  buf[256]; /* for xvgr title */
  int   slice, n;
  real  ddd;

  sprintf(buf,"Partial densities");
  if (bNumber)
    den = xvgropen(afile, buf, "Box (nm)","Density (atoms/nm\\S3\\N)");
  else if (bCount)
    den = xvgropen(afile, buf, "Slice","Absolute numbers");
  else
    den = xvgropen(afile, buf, "Box (nm)","Density (kg/m\\S3\\N)");

  xvgr_legend(den,nr_grps,grpname);

  for (slice = 0; (slice < nslices); slice++) { 
    fprintf(den,"%12g  ", slice * slWidth);
    for (n = 0; (n < nr_grps); n++) {
      if (bSymmetrize)
	ddd = (slDensity[n][slice]+slDensity[n][nslices-slice-1])*0.5;
      else
	ddd = slDensity[n][slice];
      if (bNumber)
	fprintf(den,"   %12g", ddd/1.66057);
      else
	fprintf(den,"   %12g", ddd);
    }
    fprintf(den,"\n");
  }

  fclose(den);
}
 
int main(int argc,char *argv[])
{
  static char *desc[] = {
    "Compute lipid bilayer thickness from the positions of all lipid phosphorus",
    "atoms of type LP."
  };
  
  static int  axis = 2;          /* normal to memb. default z  */
  static char *axtitle="Z"; 
  static int  ngrps   = 1;       /* nr. of groups              */
  t_pargs pa[] = {
    { "-d", FALSE, etSTR, {&axtitle}, 
      "Take the normal on the membrane in direction X, Y or Z." },
    { "-ng",       FALSE, etINT, {&ngrps},
      "Number of groups to compute thicknesses of" }
  };

  static char *bugs[] = {
    "None that I know of."
  };
  
  int nframe;            /* number of frames used      */
  real **thickness;      /* thickness per time         */
  real *time;            /* dependent variable         */
  char **grpname;        /* groupnames                 */
  int  *ngx;             /* sizes of groups            */
  t_topology *top;       /* topology 		       */ 
  atom_id   **index;     /* indices for all groups     */
  
  t_filenm  fnm[] = {    /* files for g_density 	  */
    { efTRX, "-f", NULL,  ffREAD },  
    { efNDX, NULL, NULL,  ffOPTRD }, 
    { efTPX, NULL, NULL,  ffREAD },    	    
    { efXVG,"-o","thickness",ffWRITE }, 	    
  };
  
#define NFILE asize(fnm)

  CopyRight(stderr,argv[0]);

  parse_common_args(&argc,argv,PCA_CAN_VIEW | PCA_CAN_TIME | PCA_BE_NICE,
		    NFILE,fnm,asize(pa),pa,asize(desc),desc,asize(bugs),bugs);

  /* Calculate axis */
  axis = toupper(axtitle[0]) - 'X';
  
  top = read_top(ftp2fn(efTPX,NFILE,fnm));     /* read topology file */

  snew(grpname,ngrps);
  snew(index,ngrps);
  snew(ngx,ngrps);
 
  get_index(&top->atoms,ftp2fn_null(efNDX,NFILE,fnm),ngrps,ngx,index,grpname); 

  calc_thickness(ftp2fn(efTRX,NFILE,fnm), index, ngx, &thickness, &time, 
                   &nframe, top, axis, ngrps);

  plot_thickness(thickness, time, opt2fn("-o",NFILE,fnm), nframe, ngrps);

  do_view(opt2fn("-o",NFILE,fnm), "-nxy");       /* view xvgr file */
  thanx(stderr);
  return 0;
}



